/*
 * Copyright © 2005 Novell, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Novell, Inc. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Novell, Inc. makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * NOVELL, INC. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL NOVELL, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: David Reveman <davidr@novell.com>
 */

/*
 * source HEAVILY modified by Quinn, released under GPL
 */

#define _GNU_SOURCE
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include <beryl-settings.h>
#include <beryl.h>

#include <glib.h>
#include <glib/gprintf.h>
#include <glib/gstdio.h>
#include <errno.h>
#include <signal.h>

#define KEY_CHANGE_TIMEOUT 250

static int displayPrivateIndex;
enum _foo
{
    CSM_DISPLAY_OPTION_RELOAD,
    CSM_DISPLAY_OPTION_NUM
};
typedef struct _CSMDisplay
{
    int screenPrivateIndex;

    BerylSettingsContext * context;

    CompTimeoutHandle timeoutHandle;

    InitPluginForDisplayProc initPluginForDisplay;
    SetDisplayOptionProc setDisplayOption;
    SetDisplayOptionForPluginProc setDisplayOptionForPlugin;
} CSMDisplay;

typedef struct _CSMScreen
{
    InitPluginForScreenProc initPluginForScreen;
    SetScreenOptionProc setScreenOption;
    SetScreenOptionForPluginProc setScreenOptionForPlugin;
} CSMScreen;

static gboolean reloadPending=FALSE;

#define GET_CSM_DISPLAY(d)				      \
    ((CSMDisplay *) (d)->privates[displayPrivateIndex].ptr)

#define CSM_DISPLAY(d)		     \
    CSMDisplay *gd = GET_CSM_DISPLAY (d)

#define GET_CSM_SCREEN(s, gd)				         \
    ((CSMScreen *) (s)->privates[(gd)->screenPrivateIndex].ptr)

#define CSM_SCREEN(s)						           \
    CSMScreen *gs = GET_CSM_SCREEN (s, GET_CSM_DISPLAY (s->display))

static Bool
setSetDisplayOption (CompDisplay * d, char *name, CompOptionValue * value)
{
    Bool status;

    CSM_DISPLAY (d);

    UNWRAP (gd, d, setDisplayOption);
    status = (*d->setDisplayOption) (d, name, value);
    WRAP (gd, d, setDisplayOption, setSetDisplayOption);

    if (status)
    {
        beryl_settings_context_comp_set_option_value(gd->context,NULL,name,FALSE,value);
        beryl_settings_context_write(gd->context);
    }

    return status;
}

static Bool
setSetDisplayOptionForPlugin (CompDisplay * d,
                              char *plugin,
                              char *name, CompOptionValue * value)
{
    Bool status;

    CSM_DISPLAY (d);

    UNWRAP (gd, d, setDisplayOptionForPlugin);
    status = (*d->setDisplayOptionForPlugin) (d, plugin, name, value);
    WRAP (gd, d, setDisplayOptionForPlugin, setSetDisplayOptionForPlugin);

    if (status)
    {
        beryl_settings_context_comp_set_option_value(gd->context,plugin,name,FALSE,value);
        beryl_settings_context_write(gd->context);
    }

    return status;
}

static Bool
setSetScreenOption (CompScreen * s, char *name, CompOptionValue * value)
{
    Bool status;

    CSM_SCREEN (s);
    CSM_DISPLAY(s->display);

    UNWRAP (gs, s, setScreenOption);
    status = (*s->setScreenOption) (s, name, value);
    WRAP (gs, s, setScreenOption, setSetScreenOption);

    if (status)
    {
        beryl_settings_context_comp_set_option_value(gd->context,NULL,name,TRUE,value);
        beryl_settings_context_write(gd->context);
    }

    return status;
}

static Bool
setSetScreenOptionForPlugin (CompScreen * s,
                             char *plugin,
                             char *name, CompOptionValue * value)
{
    Bool status;

    CSM_SCREEN (s);
    CSM_DISPLAY(s->display);

    UNWRAP (gs, s, setScreenOptionForPlugin);
    status = (*s->setScreenOptionForPlugin) (s, plugin, name, value);
    WRAP (gs, s, setScreenOptionForPlugin, setSetScreenOptionForPlugin);

    if (status)
    {
        beryl_settings_context_comp_set_option_value(gd->context,plugin,name,TRUE,value);
        beryl_settings_context_write(gd->context);
    }

    return status;
}

static Bool
setInitPluginForDisplay (CompPlugin * p, CompDisplay * d)
{
    Bool status;

    CSM_DISPLAY (d);

    UNWRAP (gd, d, initPluginForDisplay);
    status = (*d->initPluginForDisplay) (p, d);
    WRAP (gd, d, initPluginForDisplay, setInitPluginForDisplay);

    beryl_settings_context_comp_add_plugin(gd->context,p->vTable->name);

    if (status && p->vTable->getDisplayOptions)
    {
        CompOption *option;
        CompOptionValue value;
        int nOption;

        option = (*p->vTable->getDisplayOptions) (d, &nOption);
        while (nOption--)
            beryl_settings_context_comp_add_option(gd->context,
                    p->vTable->name,option++,FALSE);
        beryl_settings_context_read(gd->context);
        option = (*p->vTable->getDisplayOptions) (d, &nOption);
        while (nOption--)
        {
            memcpy(&value,&option->value,sizeof(CompOptionValue));
            if (beryl_settings_context_comp_get_option_value(gd->context,
                    p->vTable->name,option->name,FALSE,&value))
                d->setDisplayOptionForPlugin (d, p->vTable->name, option->name, &value);
            option++;
        }
    }

    return status;
}

static Bool
setInitPluginForScreen (CompPlugin * p, CompScreen * s)
{
    Bool status;

    CSM_SCREEN (s);
    CSM_DISPLAY(s->display);

    UNWRAP (gs, s, initPluginForScreen);
    status = (*s->initPluginForScreen) (p, s);
    WRAP (gs, s, initPluginForScreen, setInitPluginForScreen);

    beryl_settings_context_comp_add_plugin(gd->context,p->vTable->name);

    if (status && p->vTable->getScreenOptions)
    {
        CompOption *option;
        CompOptionValue value;
        int nOption;

        option = (*p->vTable->getScreenOptions) (s, &nOption);
        while (nOption--)
            beryl_settings_context_comp_add_option(gd->context,
                    p->vTable->name,option++,TRUE);
        beryl_settings_context_read(gd->context);
        option = (*p->vTable->getScreenOptions) (s, &nOption);
        while (nOption--)
        {
            memcpy(&value,&option->value,sizeof(CompOptionValue));
            if (beryl_settings_context_comp_get_option_value(gd->context,
                    p->vTable->name,option->name,TRUE,&value))
                s->setScreenOptionForPlugin (s, p->vTable->name, option->name, &value);
            option++;
        }
    }

    return status;
}

static void
reload_all (CompDisplay * d)
{
    CompOption * option;
    CompOptionValue value;
    int nOption;
    CompPlugin * p;
    CSM_DISPLAY(d);
    beryl_settings_context_read(gd->context);
    puts("Reloading all options.");
    option = compGetDisplayOptions (d, &nOption);
    while (nOption--)
    {
        memcpy(&value,&option->value,sizeof(CompOptionValue));
        if (beryl_settings_context_comp_get_option_value(
                    gd->context,NULL,option->name,FALSE,&value))
            d->setDisplayOption(d,option->name,&value);
        option++;
    }
    option = compGetScreenOptions (d->screens, &nOption);
    while (nOption--)
    {
        memcpy(&value,&option->value,sizeof(CompOptionValue));
        if (beryl_settings_context_comp_get_option_value(
                    gd->context,NULL,option->name,TRUE,&value))
            d->screens->setScreenOption(d->screens,option->name,&value);
        option++;
    }
    for(p=getPlugins();p;p=p->next)
    {
        if (!strcmp(p->vTable->name,"settings"))
            continue;
        if (p->vTable->getDisplayOptions)
        {
            option = p->vTable->getDisplayOptions (d, &nOption);
            while (nOption--)
            {
                memcpy(&value,&option->value,sizeof(CompOptionValue));
                if (beryl_settings_context_comp_get_option_value(
                            gd->context,p->vTable->name,option->name,FALSE,&value))
                    d->setDisplayOptionForPlugin(d,p->vTable->name,option->name,&value);
                option++;
            }
        }
        if (p->vTable->getScreenOptions)
        {
            option = p->vTable->getScreenOptions (d->screens, &nOption);
            while (nOption--)
            {
                memcpy(&value,&option->value,sizeof(CompOptionValue));
                if (beryl_settings_context_comp_get_option_value(
                            gd->context,p->vTable->name,option->name,TRUE,&value))
                    d->screens->setScreenOptionForPlugin(d->screens,p->vTable->name,option->name,&value);
                option++;
            }
        }
    }
}

static void
set_reload_all (int sig)
{
    if (sig == SIGUSR1)
    {
        reloadPending=TRUE;
    }
}

static Bool
setTimeout (void *closure)
{
    CompDisplay * d = closure;

    while (g_main_pending ())
        g_main_iteration (FALSE);

    if (reloadPending)
    {
        reloadPending=FALSE;
        reload_all(d);
    }

    return TRUE;
}

static Bool
setInitDisplay (CompPlugin * p, CompDisplay * d)
{
    CompOption *option;
    int nOption;
    CSMDisplay *gd;

    gd = malloc (sizeof (CSMDisplay));
    if (!gd)
        return FALSE;

    gd->screenPrivateIndex = allocateScreenPrivateIndex (d);
    if (gd->screenPrivateIndex < 0)
    {
        free (gd);
        return FALSE;
    }

   d->privates[displayPrivateIndex].ptr = gd;

    gd->context = beryl_settings_context_comp_new();
    beryl_settings_context_comp_add_plugin(gd->context,NULL);

    option = compGetDisplayOptions (d, &nOption);
    while (nOption--)
    {
        beryl_settings_context_comp_add_option(gd->context,NULL,option++,FALSE);
    }
    beryl_settings_context_read(gd->context);
    option = compGetDisplayOptions (d, &nOption);
    while (nOption--)
    {
        CompOptionValue value;
        memcpy(&value,&option->value,sizeof(CompOptionValue));
        if (beryl_settings_context_comp_get_option_value(
                    gd->context,NULL,option->name,FALSE,&value))
            d->setDisplayOption(d,option->name,&value);
        option++;
    }

    WRAP (gd, d, initPluginForDisplay, setInitPluginForDisplay);
    WRAP (gd, d, setDisplayOption, setSetDisplayOption);
    WRAP (gd, d, setDisplayOptionForPlugin, setSetDisplayOptionForPlugin);
 
    gd->timeoutHandle = compAddTimeout (KEY_CHANGE_TIMEOUT, setTimeout, d);

    signal(SIGUSR1, set_reload_all);
    reloadPending = TRUE;

    return TRUE;
}

static void
setFiniDisplay (CompPlugin * p, CompDisplay * d)
{
    CSM_DISPLAY (d);

    signal(SIGUSR1, SIG_DFL);

    beryl_settings_context_destroy(gd->context);
    compRemoveTimeout (gd->timeoutHandle);

    freeScreenPrivateIndex (d, gd->screenPrivateIndex);

    UNWRAP (gd, d, initPluginForDisplay);
    UNWRAP (gd, d, setDisplayOption);
    UNWRAP (gd, d, setDisplayOptionForPlugin);

    free (gd);
}

static Bool
setInitScreen (CompPlugin * p, CompScreen * s)
{
    CompOption *option;
    int nOption;
    CSMScreen *gs;

    CSM_DISPLAY (s->display);

    gs = malloc (sizeof (CSMScreen));
    if (!gs)
        return FALSE;

    WRAP (gs, s, initPluginForScreen, setInitPluginForScreen);
    WRAP (gs, s, setScreenOption, setSetScreenOption);
    WRAP (gs, s, setScreenOptionForPlugin, setSetScreenOptionForPlugin);

    s->privates[gd->screenPrivateIndex].ptr = gs;

    beryl_settings_context_comp_add_plugin(gd->context,NULL);

    option = compGetScreenOptions (s, &nOption);
    while (nOption--)
    {
        beryl_settings_context_comp_add_option(gd->context,NULL,option++,TRUE);
    }
    beryl_settings_context_read(gd->context);
    option = compGetScreenOptions (s, &nOption);
    while (nOption--)
    {
        CompOptionValue value;
        memcpy(&value,&option->value,sizeof(CompOptionValue));
        if (beryl_settings_context_comp_get_option_value(
                    gd->context,NULL,option->name,TRUE,&value))
            s->setScreenOption(s,option->name,&value);
        option++;
    }

    return TRUE;
}

static void
setFiniScreen (CompPlugin * p, CompScreen * s)
{
    CSM_SCREEN (s);

    UNWRAP (gs, s, initPluginForScreen);
    UNWRAP (gs, s, setScreenOption);
    UNWRAP (gs, s, setScreenOptionForPlugin);

    free (gs);
}


static Bool
setInit (CompPlugin * p)
{
    displayPrivateIndex = allocateDisplayPrivateIndex ();
    if (displayPrivateIndex < 0)
        return FALSE;

    return TRUE;
}

static void
setFini (CompPlugin * p)
{
    if (displayPrivateIndex >= 0)
        freeDisplayPrivateIndex (displayPrivateIndex);
}

CompPluginVTable setVTable = {
    "settings",
    "Settings Manager",
    "Use beryl-settings to configure me!",
    setInit,
    setFini,
    setInitDisplay,
    setFiniDisplay,
    setInitScreen,
    setFiniScreen,
    0,                          /* InitWindow */
    0,                          /* FiniWindow */
    0,                          /* GetDisplayOptions */
    0,                          /* SetDisplayOption */
    0,                          /* GetScreenOptions */
    0,                          /* SetScreenOption */
    0,
    0,
    0,
    0,
    BERYL_ABI_INFO,
    "beryl-plugins"
};

CompPluginVTable *
getCompPluginInfo (void)
{
    return &setVTable;
}
